"""For handling the cost matrix for chemical graph edit distance."""

import abc

import networkx as nx
import numpy as np
import numpy.typing as npt
from networkx import Graph
from networkx.classes.coreviews import AtlasView


class ChemicalGEDCostMatrix(abc.ABC):
    """
    Abstract base class for chemical graph edit distance cost matrix

    These Matrices are used to determine the cost of any edit operation.

    The costs of node substitution, insertion and deletion costs
    need to be defined. These should return a cost matrix (for substitution)
    or a cost vector (for insertion and deletion) of the nodes.
    This is done by the functions
    - `get_node_substitution_costs`
    - `get_node_insertion_costs`
    - `get_node_deletion_costs`

    Likewise, the cost functions for edge substitution, insertion and deletion
    are needed. These are a bit different. Rather than operating on the whole graph,
    they operate locally on the nodes. They should return a cost matrix (substitution)
    or a cost vector (insertion and deletion) of the edges *for the two given nodes*
    only. So if we have two nodes n1 (from graph1) and n2 (from graph2), and n1 has
    a degree of 3 (3 bonds) and n2 has a degree of 2 (2 bonds), then the substitution
    cost matrix should be of size 3x2, where the element at position (i, j) is the
    substitution cost of the i-th bond of n1 with the j-th bond of n2. Likewise,
    the deletion/insertion cost vector should be the same size as the respective node's degree.

    - `get_edge_substitution_costs`
    - `get_edge_insertion_costs`
    - `get_edge_deletion_costs`
    """

    @abc.abstractmethod
    def get_node_substitution_costs(self, g1: Graph, g2: Graph):
        """
        Get the node substitution costs as a matrix between two ordered graphs.

        This method should return a cost matrix where the element at position (i, j)
        represents the cost of substituting node i in n1 with node j in n2.

        The node order in these graphs *MUST* stay static over the object lifetime
        (at least during GED calculation),
        otherwise the generated cost matrix may become invalid during calculation.

        Parameters
        ----------
        g1 : Graph
            The first chemical graph.
        g2 : Graph
            The second chemical graph.

        Returns
        -------
        np.ndarray
            A 2D cost matrix for node substitutions.
            shape should be (len(n1.nodes), len(n2.nodes))
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_edge_substitution_costs(self, n1: AtlasView, n2: AtlasView):
        """
        Get the edge substitution costs as a matrix between two nodes.

        This method should return a cost matrix where the element at position (i, j)
        represents the cost of substituting edge i in n1 with edge j in n2.

        The edge order in these graphs *MUST* stay static over the object lifetime
        (at least during GED calculation),
        otherwise the generated cost matrix may become invalid during calculation.

        Notes
        -----
        You can get the adjacency list of a node in a graph from network x using
        graph.adj[n] OR using g[n] where n is the node key/index. These are called
        `AtlasViews` in NetworkX and can be treated as dictionaries for the most part.

        Parameters
        ----------
        n1 : AtlasView
            The adjacency list of the first chemical graph node.
        n2 : AtlasView
            The adjacency list of the first chemical graph node.

        Returns
        -------
        np.ndarray
            A 2D cost matrix for edge substitutions.
            shape should be (degree(n1), degree(n2))
        """
        raise NotImplementedError

    def get_node_insertion_costs(self, g: Graph):
        """
        Get the node insertion costs as a vector for a graph.

        This method should return a cost vector where the element at position i
        represents the cost of inserting node i in g.

        Parameters
        ----------
        g : Graph
            The chemical graph.

        Returns
        -------
        np.ndarray
            A 1D cost vector for node insertions.
            shape should be (len(g.nodes),)
        """
        raise NotImplementedError

    def get_node_deletion_costs(self, g: Graph):
        """
        Get the node deletion costs as a vector for a graph.

        This method should return a cost vector where the element at position i
        represents the cost of deleting node i in g.

        Parameters
        ----------
        g : Graph
            The chemical graph.

        Returns
        -------
        np.ndarray
            A 1D cost vector for node deletions.
            shape should be (len(g.nodes),)
        """
        raise NotImplementedError

    def get_edge_insertion_costs(self, n: AtlasView):
        """
        Get the edge insertion costs as a vector for a node in a graph.

        This method should return a cost vector where the element at position i
        represents the cost of inserting edge i in the node n.

        Parameters
        ----------
        n : AtlasView
            The adjacency list of the chemical graph node.

        Returns
        -------
        np.ndarray
            A 1D cost vector for edge insertions.
            shape should be (degree(n),)
        """
        raise NotImplementedError

    def get_edge_deletion_costs(self, n: AtlasView):
        """
        Get the edge deletion costs as a vector for a node in a graph.

        This method should return a cost vector where the element at position i
        represents the cost of deleting edge i in the node n.

        Parameters
        ----------
        n : AtlasView
            The adjacency list of the chemical graph node.

        Returns
        -------
        np.ndarray
            A 1D cost vector for edge deletions.
            shape should be (degree(n),)
        """
        raise NotImplementedError


class UniformElementCostMatrix(ChemicalGEDCostMatrix):
    """
    A uniform cost matrix for chemical graph edit distance.

    This cost matrix assumes all substitutions, insertions, and deletions of atoms
    are worth a cost of 1, not matter which type the atom is *or* what type the
    atom was swapped into. The same goes for bonds, where all bond substitutions,
    insertions, and deletions are worth a cost of 1, regardless of the bond type.

    This is good for if you want to measure GED in a more "pure" sense, where
    we are just asking how much the graph representing a chemical has changed.
    If you want to account for chemical properties, you should write a more
    advanced cost matrix that takes into account the atom and bond types.
    For example, minimize cost for O to N substitutions if they are both
    Hydrogen bond donors, but increase the cost for O to C substitutions.

    Parameters
    ----------
    node_sub_cost : float, default=1.0
        The uniform cost of substituting a node (atom).
    node_ins_cost : float, default=1.0
        The uniform cost of inserting a node (atom).
    node_del_cost : float, default=1.0
        The uniform cost of deleting a node (atom).
    edge_sub_cost : float, default=1.0
        The uniform cost of substituting an edge (bond).
    edge_ins_cost : float, default=1.0
        The uniform cost of inserting an edge (bond).
    edge_del_cost : float, default=1.0
        The uniform cost of deleting an edge (bond).
    """

    def __init__(
        self,
        node_sub_cost=1.0,
        node_ins_cost=1.0,
        node_del_cost=1.0,
        edge_sub_cost=1.0,
        edge_ins_cost=1.0,
        edge_del_cost=1.0,
    ):
        """Initialize an object"""
        self.node_sub_cost = node_sub_cost
        self.node_ins_cost = node_ins_cost
        self.node_del_cost = node_del_cost

        self.edge_sub_cost = edge_sub_cost
        self.edge_ins_cost = edge_ins_cost
        self.edge_del_cost = edge_del_cost

    def get_node_substitution_costs(self, g1: Graph, g2: Graph) -> npt.NDArray:
        """
        Get the node substitution costs as a matrix between two ordered graphs.

        This method returns a cost matrix where the element at position (i, j)
        represents the cost of substituting node i in n1 with node j in n2.

        It uses only the element type to determine if two nodes are the same,
        if they are the cost is 0, otherwise it is `self.node_sub_cost`.

        Parameters
        ----------
        g1 : Graph
            The first chemical graph.
        g2 : Graph
            The second chemical graph.

        Returns
        -------
        np.ndarray
            A 2D cost matrix for node substitutions.
            shape should be (len(n1.nodes), len(n2.nodes))
        """
        # all the atomic numbers of the nodes in the graphs
        v1 = list(nx.get_node_attributes(g1, "atomic_num").values())
        v2 = list(nx.get_node_attributes(g2, "atomic_num").values())

        # generate the cost matrix
        return ~np.equal.outer(v1, v2) * self.node_sub_cost

    def get_node_insertion_costs(self, g: Graph) -> npt.NDArray:
        """
        Get the node insertion costs as a vector for a graph.

        Will return a vector of the same length as the number of nodes in the graph,
        filled with the value in `self.node_ins_cost.

        Parameters
        ----------
        g: Graph
            the chemical graph to get the insertion costs for

        Returns
        -------
        np.ndarray
            A 1D cost vector for node insertions.
            shape should be (len(g.nodes),)
        """
        return np.full(len(g), self.node_ins_cost)

    def get_node_deletion_costs(self, g: Graph) -> npt.NDArray:
        """
        Get the node deletion costs as a vector for a graph.

        Will return a vector of the same length as the number of nodes in the graph,
        filled with the value in `self.node_del_cost.

        Parameters
        ----------
        g: Graph
            the chemical graph to get the deletion costs for

        Returns
        -------
        np.ndarray
            A 1D cost vector for node deletions.
            shape should be (len(g.nodes),)
        """
        return np.full(len(g), self.node_del_cost)

    def get_edge_substitution_costs(self, n1: AtlasView, n2: AtlasView) -> npt.NDArray:
        """
        Get the edge substitution costs as a matrix between two nodes.

        This method returns a cost matrix where the element at position (i, j)
        represents the cost of substituting edge i in n1 with edge j in n2.

        It uses only the bond type to determine if two edges are the same,
        if they are the cost is 0, otherwise it is `self.edge_sub_cost`.

        Parameters
        ----------
        n1 : AtlasView
            The adjacency list of the first chemical graph node.
        n2 : AtlasView
            The adjacency list of the second chemical graph node.

        Returns
        -------
        np.ndarray
            A 2D cost matrix for edge substitutions.
            shape should be (degree(n1), degree(n2))
        """
        # get the bond types of the edges in the nodes
        e1 = [e["bond_type"] for e in n1.values()]
        e2 = [e["bond_type"] for e in n2.values()]

        # generate the cost matrix
        return ~np.equal.outer(e1, e2) * self.edge_sub_cost

    def get_edge_insertion_costs(self, n: AtlasView) -> npt.NDArray:
        """
        Get the edge insertion costs as a vector for a node in a graph.

        Will return a vector of the same length as the number of edges in the node,
        filled with the value in `self.edge_ins_cost`.

        Parameters
        ----------
        n : AtlasView
            The adjacency list of the chemical graph node.

        Returns
        -------
        np.ndarray
            A 1D cost vector for edge insertions.
            shape should be (degree(n),)
        """
        return np.full(len(n), self.edge_ins_cost)

    def get_edge_deletion_costs(self, n: AtlasView) -> npt.NDArray:
        """
        Get the edge deletion costs as a vector for a node in a graph.

        Will return a vector of the same length as the number of edges in the node,
        filled with the value in `self.edge_del_cost`.

        Parameters
        ----------
        n : AtlasView
            The adjacency list of the chemical graph node.

        Returns
        -------
        np.ndarray
            A 1D cost vector for edge deletions.
            shape should be (degree(n),)
        """
        return np.full(len(n), self.edge_del_cost)



"""handles calculation of approximate chemical GEDs"""

import itertools
from typing import Literal, Sequence, overload

import numpy as np
import numpy.typing as npt
from networkx import Graph
from networkx.classes.coreviews import AtlasView
from scipy.optimize import linear_sum_assignment
from tqdm import tqdm

from .chem_utils import Molable, mol_to_nx, to_mol
from .cost import ChemicalGEDCostMatrix


MAX_INT32 = 2147483647


class ApproximateChemicalGED:
    r"""
    Calculator for approximate graph edit distance (GED) between two chemical graphs.

    Parameters
    ----------
    cost_matrix : ChemicalGEDCostMatrix
        the cost matrix that defines the costs of node and edge edit operations
        on the chemical graphs.

    Notes
    -----
    This implements a bipartite graph matching algorithm to approximate GED[1]_.

    The algorithm requires a way to compute the cost matrix of node and edge edit operations
    between two graphs. The cost matrix between two graphs is defined square matrix of size
    number of nodes in graph 1 + number of nodes in graph 2.

    .. math::

    \\mathbf{C} =
    \\begin{bmatrix}
    c_{1,1} & c_{1,2} & \\cdots & c_{1,m} & c_{1,\\epsilon} & \\infty & \\cdots & \\infty \\\\
    c_{2,1} & c_{2,2} & \\cdots & c_{2,m} & \\infty & c_{2,\\epsilon} & \\ddots & \\vdots \\\\
    \\vdots & \\vdots & \\ddots & \\vdots & \\vdots & \\ddots & \\ddots & \\infty \\\\
    c_{n,1} & c_{n,2} & \\cdots & c_{n,m} & \\infty & \\cdots & \\infty & c_{n,\\epsilon}\\\\
    c_{\\epsilon, 1} & \\infty & \\cdots & \\infty & 0 & 0 & \\cdots & 0 \\\\
    \\infty & c_{\\epsilon, 2} & \\ddots & \\vdots & 0 & 0 & \\ddots & \\vdots \\\\
    \\vdots & \\ddots & \\ddots & \\infty & \\vdots & \\ddots & \\ddots & 0 \\\\
    \\infty & \\cdots & \\infty & c_{\\epsilon, m} & 0 & \\cdots & 0 & 0 \\\\
    \\end{bmatrix}

    where :math:`c_{n,m}` is the cost of substituting node :math:`n` in graph :math:`n1`
    with node :math:`m` in graph :math:`n2`, :math:`c_{n,\\epsilon}` is the cost of deleting
    a node from :math:`n1`, and :math:`c_{\\epsilon,m}` is the cost of inserting node
    :math:`m` from :math:`n2` into :math:`n1`.

    Nodes can only be deleted or inserted once, so only the diagonals are populated in the
    insertion (bottom left) and deletion (upper right) sections of the cost matrix.
    The cost of doing another insertion or deletion is set to infinity, as it is not allowed.
    Particularly, we set it to the largest integer value to avoid numerical issues with the
    implementation of Munkres algorithm used by SciPy.
    The bottom right is always 0, as it represents the cost of substituting a missing node with
    another missing node, which has no cost.

    Then the Munkres algorithm is applied to find the optimal assignment of nodes and edges
    between the two graphs that minimizes the total cost of the edit operations.
    These costs are then summed to approximate the graph edit distance

    References
    ----------
    .. [1] Neuhaus, C., & Bunke, H. (2007). A graph edit distance
           based on the maximal common subgraph with application to
           structural document image recognition. In Document Analysis
           Systems VII (pp. 188-199). IEEE.
    """

    def __init__(self, cost_matrix: ChemicalGEDCostMatrix):
        self.cost_matrix = cost_matrix

    def _edge_cost_matrix(self, n1: AtlasView, n2: AtlasView):
        """
        Given two nodes, generate the cost matrix for edge edit operations.

        Parameters
        ----------
        n1: AtlasView
            the adj info the node from the first graph
        n2: AtlasView
            the adj info the node from the second graph

        Returns
        -------
        np.ndarray
            the cost matrix for edge edit operations between the two nodes
            shape will be (degree(n1) + degree(n2), degree(n1) + degree(n2))
        """
        cost_matrix = np.zeros([len(n1) + len(n2), len(n1) + len(n2)])

        # Insertion
        cost_matrix[len(n1) :, 0 : len(n2)] = MAX_INT32
        np.fill_diagonal(
            cost_matrix[len(n1) :, 0 : len(n2)], self.cost_matrix.get_edge_insertion_costs(n1)
        )

        # Deletion
        cost_matrix[0 : len(n1), len(n2) :] = MAX_INT32
        np.fill_diagonal(
            cost_matrix[0 : len(n1), len(n2) :], self.cost_matrix.get_edge_deletion_costs(n2)
        )

        # Substitution
        cost_matrix[0 : len(n1), 0 : len(n2)] = self.cost_matrix.get_edge_substitution_costs(
            n1, n2
        )

        return cost_matrix

    def _edge_edit_cost(self, n1: AtlasView, n2: AtlasView):
        """
        Compute the approximate graph edit cost for edges between two nodes.

        This is the part of the algorithm that attempts to match the local structures
        to avoid a much more expensive global edit distance calculation.

        Parameters
        ----------
        n1: AtlasView
            the adj info the node from the first graph
        n2: AtlasView
            the adj info the node from the second graph

        Returns
        -------
        float
            the approximate graph edit cost for edges between the two nodes
        """
        # Compute cost matrix
        cost_matrix = self._edge_cost_matrix(n1, n2)

        # Munkres algorithm
        row_ind, col_ind = linear_sum_assignment(cost_matrix)

        # Edge edit distance between the two nodes
        dist = cost_matrix[row_ind, col_ind].sum()

        return dist

    def _get_full_cost_matrix(self, g1: Graph, g2: Graph):
        """
        Compute the full cost matrix for the two graphs, including node and edge edit operations.

        This matrix is what is used to compute the approximate graph edit distance with
        the Munkres algorithm.

        Parameters
        ----------
        g1: Graph
            the first graph
        g2: Graph
            the second graph

        Returns
        -------
        np.ndarray
            the full cost matrix for the two graphs, shape will be
            (len(g1) + len(g2), len(g1) + len(g2)) where len(g) is
            the number of nodes in the graph.
        """
        cost_matrix = np.zeros([len(g1) + len(g2), len(g1) + len(g2)])

        # Insertion
        cost_matrix[len(g1) :, 0 : len(g2)] = MAX_INT32
        np.fill_diagonal(
            cost_matrix[len(g1) :, 0 : len(g2)],
            self.cost_matrix.get_node_insertion_costs(g1)
            + [self.cost_matrix.get_edge_insertion_costs(g1[i]).sum() for i in range(len(g1))],
        )

        # Deletion
        cost_matrix[0 : len(g1), len(g2) :] = MAX_INT32
        np.fill_diagonal(
            cost_matrix[0 : len(g1), len(g2) :],
            self.cost_matrix.get_node_deletion_costs(g2)
            + [self.cost_matrix.get_edge_deletion_costs(g2[i]).sum() for i in range(len(g2))],
        )

        # Substitution
        node_dist = self.cost_matrix.get_node_substitution_costs(g1, g2)
        # update the substitution costs with the edge edit scores
        for i, n in enumerate(g1.nodes()):
            for j, m in enumerate(g2.nodes()):
                node_dist[i, j] += self._edge_edit_cost(g1[n], g2[m])
        cost_matrix[0 : len(g1), 0 : len(g2)] = node_dist

        return cost_matrix

    @overload
    def _ged(
        self, g1: Graph, g2: Graph, return_assignment: Literal[False]
    ) -> tuple[float, None]: ...

    @overload
    def _ged(
        self, g1: Graph, g2: Graph, return_assignment: Literal[True]
    ) -> tuple[float, tuple[npt.NDArray, npt.NDArray]]: ...

    @overload
    def _ged(
        self, g1: Graph, g2: Graph, return_assignment: bool
    ) -> tuple[float, tuple[npt.NDArray, npt.NDArray] | None]: ...

    def _ged(
        self, g1: Graph, g2: Graph, return_assignment: bool = False
    ) -> tuple[float, tuple[npt.NDArray, npt.NDArray] | None]:
        """
        Compute the approximate graph edit distance between two chemicals

        Parameters
        ----------
        g1: Graph
            the first chemical graph
        g2: Graph
            the second chemical graph

        Returns
        -------
        distance, assignment: float, tuple[npt.ndarray, npt.ndarray] | None
            'distance' is the approximate graph edit distance between the two graphs,
            'assignment' is a tuple of two numpy arrays defining the assignment
            between the nodes of the two graphs
        """
        # Compute cost matrix
        cost_matrix = self._get_full_cost_matrix(g1, g2)

        # Munkres algorithm
        row_ind, col_ind = linear_sum_assignment(cost_matrix)

        # Graph edit distance
        dist = cost_matrix[row_ind, col_ind].sum()

        if return_assignment:
            not_assign = np.invert((row_ind >= len(g1)) * (col_ind >= len(g2)))
            return dist, (row_ind[not_assign], col_ind[not_assign])
        else:
            return dist, None

    @overload
    def compute_ged(
        self, chemical1: Molable, chemical2: Molable, return_assignment: Literal[False]
    ) -> float: ...

    @overload
    def compute_ged(
        self, chemical1: Molable, chemical2: Molable, return_assignment: Literal[True]
    ) -> tuple[float, tuple[npt.NDArray, npt.NDArray]]: ...

    @overload
    def compute_ged(
        self, chemical1: Molable, chemical2: Molable, return_assignment: bool
    ) -> float | tuple[float, tuple[npt.NDArray, npt.NDArray]]: ...

    def compute_ged(
        self, chemical1: Molable, chemical2: Molable, return_assignment: bool = False
    ) -> float | tuple[float, tuple[npt.NDArray, npt.NDArray]]:
        """
        Compute the approximate graph edit distance between two chemicals

        Will convert the chemicals to rdkit.Mol objects if they are not already,
        and then covert them into NetworkX Graph objects. See `mol_to_nx` for more details.

        NOTE: it is inefficient to call this function multiple times with the same chemicals,
        as the chemicals will be converted to NetworkX Graph objects each time. If you have
        a set of chemcials that you want to get a pairwise distance for, use the `pdist``
        or `cdist` functions instead.

        Parameters
        ----------
        chemical1: Molable (str or rdkit.Mol)
            the first chemical to compare, can be a SMILES string or rdkit.Mol object
        chemical2: Molable (str or rdkit.Mol)
            the second chemical to compare, can be a SMILES string or rdkit.Mol object
        return_assignment: bool, default=False
            whether to return the assignment of nodes between the two graphs.
            If True, will return a tuple of two numpy arrays defining the assignment
            between the nodes of the two graphs. If False, will only return the distance.

        Returns
        -------
        distance, assignment: float, tuple[float, tuple[npt.ndarray, npt.ndarray]]
            'distance' is the approximate graph edit distance between the two chemicals,
            'assignment' is a tuple of two numpy arrays defining the assignment
            between the nodes of the two graphs if `return_assignment` is True.
        """
        mol1 = to_mol(chemical1, fail_on_error=True)
        mol2 = to_mol(chemical2, fail_on_error=True)

        g1: Graph = mol_to_nx(mol1)
        g2: Graph = mol_to_nx(mol2)

        dist, assignment = self._ged(g1, g2, return_assignment=return_assignment)
        if return_assignment:
            if assignment is None:
                raise ValueError("Assignment is None, this should not happen.")
            else:
                return dist, assignment
        else:
            return dist

    def pdist(self, chemicals: Sequence[Molable], use_tqdm: bool = False) -> npt.NDArray:
        """
        Given a list of chemicals, get the approximate graph edit distance between chemicals

        Notes
        -----
        This will return distances in vector-form distance vector.
        Yu can use `scipy.spatial.distance.squareform` to convert this to a square distance matrix.

        Parameters
        ----------
        chemicals: Sequence[Molable]
            a sequence of chemicals to compute the pairwise distances for.
            Each chemical can be a SMILES string or rdkit.Mol object.
        use_tqdm: bool, default=False
            enable a tqdm progress bar for the pairwise distance calculation.

        Returns
        -------
        np.ndarray
            a condensed distance vector containing the pairwise distances between the chemicals.
            The length of the vector will be n * (n - 1) / 2, where n is the number of chemicals.
        """
        mols = [
            to_mol(c, fail_on_error=True)
            for c in tqdm(
                chemicals, disable=not use_tqdm, desc="Converting chemicals to Mol objects"
            )
        ]
        graphs = [
            mol_to_nx(mol)
            for mol in tqdm(
                mols, disable=not use_tqdm, desc="Converting Mol objects to NetworkX graphs"
            )
        ]

        dists = []
        total_calcs = len(graphs) * (len(graphs) - 1) // 2
        for g1, g2 in tqdm(
            itertools.combinations(graphs, 2),
            total=total_calcs,
            disable=not use_tqdm,
            desc="Computing pairwise Approx. GED distances between chemicals",
        ):
            dists.append(self._ged(g1, g2, return_assignment=False)[0])

        return np.array(dists, dtype=np.float64)

    def cdist(
        self, chemicals1: Sequence[Molable], chemicals2: Sequence[Molable], use_tqdm: bool = False
    ) -> npt.NDArray:
        """
        Given two sequences of chemicals, compute the approximate graph edit distance between them

        Parameters
        ----------
        chemicals1: Sequence[Molable]
            a sequence of chemicals to compute the pairwise distances for.
            Each chemical can be a SMILES string or rdkit.Mol object.
        chemicals2: Sequence[Molable]
            a second sequence of chemicals to compute the pairwise distances for.
            Each chemical can be a SMILES string or rdkit.Mol object.
        use_tqdm: bool, default=False
            enable a tqdm progress bar for the pairwise distance calculation.

        Returns
        -------
        np.ndarray
            a condensed distance vector containing the pairwise distances between the chemicals.
            The length of the vector will be n * m, where n is the number of chemicals in
            `chemicals1` and m is the number of chemicals in `chemicals2`.
        """
        mols1 = [
            to_mol(c, fail_on_error=True)
            for c in tqdm(
                chemicals1,
                disable=not use_tqdm,
                desc="Converting first set of chemicals to Mol objects",
            )
        ]
        mols2 = [
            to_mol(c, fail_on_error=True)
            for c in tqdm(
                chemicals2,
                disable=not use_tqdm,
                desc="Converting second set of chemicals to Mol objects",
            )
        ]

        graphs1 = [
            mol_to_nx(mol)
            for mol in tqdm(
                mols1,
                disable=not use_tqdm,
                desc="Converting first set of Mol objects to NetworkX graphs",
            )
        ]
        graphs2 = [
            mol_to_nx(mol)
            for mol in tqdm(
                mols2,
                disable=not use_tqdm,
                desc="Converting second set of Mol objects to NetworkX graphs",
            )
        ]

        dists = []
        total_calcs = len(graphs1) * len(graphs2)
        for g1, g2 in tqdm(
            itertools.product(graphs1, graphs2),
            total=total_calcs,
            disable=not use_tqdm,
            desc="Computing pairwise Approx. GED distances between chemicals",
        ):
            dists.append(self._ged(g1, g2, return_assignment=False)[0])

        return np.array(dists, dtype=np.float64)



"""Utilities for converting between RDKit Mol objects and NetworkX Graphs."""

from typing import Literal, Union, overload

import networkx as nx
from rdkit import Chem
from rdkit.rdBase import BlockLogs


Molable = Union[str, Chem.Mol]


@overload
def to_mol(smi: Molable, fail_on_error: Literal[False]) -> Chem.Mol: ...


@overload
def to_mol(smi: Molable, fail_on_error: Literal[True]) -> Chem.Mol | None: ...


def to_mol(smi: Molable, fail_on_error: bool = True) -> Chem.Mol | None:
    """
    Given an object, attempt to convert it to a Chem.Mol object

    Notes
    -----
    Can only covert valid SMILES str and rdkit.Mol objects.
    If a rdkit.Mol object is passed, the same object will be returned.

    Parameters
    ----------
    smi: Any
        object to convert to a Chem.Mol
    fail_on_error: bool
        whether to raise an exception when converting fails.
        if True, will return None when a conversion fails

    Returns
    -------
    Chem.Mol | None
        Will return a Chem.Mol object if conversion is successful,
        if fail_on_error is True and conversion fails, will return None.

    Raises
    ------
    ValueError
        if the SMILES cannot be parsed by rdkit
    TypeError
        if the passed object is not a type that can be converted to a Chem.Mol
    """
    _lock = BlockLogs()  # this turns off the rdkit logger
    if isinstance(smi, Chem.Mol):  # return if already a Chem.Mol
        return smi
    elif isinstance(smi, str):
        _mol = Chem.MolFromSmiles(smi)
        if _mol is None:
            if fail_on_error:
                raise ValueError(f"SMILES {smi} cannot be parsed by RDKit")
        return _mol
    else:
        if fail_on_error:
            raise TypeError(f"cannot convert type {type(smi)} to type rdkit.Mol")
        else:
            return None


def mol_to_nx(mol: Chem.Mol) -> nx.Graph:
    """
    Coverts an RDKit Mol object to a NetworkX Graph.

    Will store the following attributes for each node (atom):
    - atomic_num: Atomic number of the atom
    - hybridization: Hybridization state of the atom
    - is_aromatic: Whether the atom is aromatic
    - is_in_ring: Whether the atom is part of a ring
    - hydrogen: Total number of hydrogen atoms attached to the atom
    - degree: Degree of the atom (number of bonds)

    Will save only the bond type for each edge (bond)

    See the RDKit documentation for more details on the attributes,
    all of them are functionally Enums and can be treated as integers.

    Parameters
    ----------
    mol: Chem.Mol

    Returns
    -------
    nx.Graph
    """
    graph = nx.Graph()
    max_idx = 0
    for atom in mol.GetAtoms():
        graph.add_node(
            atom.GetIdx(),
            atomic_num=atom.GetAtomicNum(),
            hybridization=atom.GetHybridization(),
            is_aromatic=atom.GetIsAromatic(),
            is_in_ring=atom.IsInRing(),
            hydrogen=atom.GetTotalNumHs(),
            degree=atom.GetDegree(),
            formal_charge=atom.GetFormalCharge(),
        )
        if atom.GetIdx() > max_idx:
            max_idx = atom.GetIdx()
    for bond in mol.GetBonds():
        graph.add_edge(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx(), bond_type=bond.GetBondType())
    return graph


def nx_to_mol(graph: nx.Graph) -> Chem.Mol:
    """
    Give a NetworkX Graph, convert it to a Chem.Mol object.

    NOTE: This is only stable for converting graphs that were created from
    the `mol_to_nx` function in this package.
    It is also lossy, so some information may be lost in the conversion for
    Mol -> Graph -> Mol.

    Parameters
    ----------
    graph: nx.Graph
        the NetworkX Graph to convert to a Chem.Mol object

    Returns
    -------
    Chem.Mol
    """
    mol = Chem.RWMol()
    atomic_nums = nx.get_node_attributes(graph, "atomic_num")
    node_is_aromatics = nx.get_node_attributes(graph, "is_aromatic")
    node_hybridization = nx.get_node_attributes(graph, "hybridization")
    node_to_idx = {}
    for node in graph.nodes():
        a = Chem.Atom(atomic_nums[node])
        a.SetIsAromatic(node_is_aromatics[node])
        a.SetHybridization(node_hybridization[node])
        idx = mol.AddAtom(a)
        node_to_idx[node] = idx

    bond_types = nx.get_edge_attributes(graph, "bond_type")
    for edge in graph.edges():
        first, second = edge
        idx_first = node_to_idx[first]
        idx_second = node_to_idx[second]
        bond_type = bond_types[first, second]
        mol.AddBond(idx_first, idx_second, bond_type)

    Chem.SanitizeMol(mol)
    return mol



"""module info for chemged"""

__version__ = "0.1.0"
__author__ = "James Wellnitz"

from .chem_utils import mol_to_nx
from .chemged import ApproximateChemicalGED
from .cost import ChemicalGEDCostMatrix, UniformElementCostMatrix


__all__ = [
    "ChemicalGEDCostMatrix",
    "UniformElementCostMatrix",
    "ApproximateChemicalGED",
    "mol_to_nx",
]



"""Plotting utilities for visualizing chemical structure assignments."""

try:
    import matplotlib.pyplot as plt
except ImportError as e:
    raise ImportError(
        "matplotlib is required for plotting. Please install it using 'pip install matplotlib'."
    ) from e

from os import PathLike
from typing import Optional

import networkx as nx
import numpy as np

from .chem_utils import Molable, mol_to_nx, to_mol


def plot_assignment(
    chemical1: Molable,
    chemical2: Molable,
    assignment,
    save_path: Optional[PathLike] = None,
    show: bool = True,
) -> "plt.Figure":
    """
    Plot the assignment of two chemical structures, mapping nodes from one to the other.

    This might be useful to visualize if the approximate GED calculator is making good
    node mapping. It could be hard to interpret the results if there are lots of atoms
    though.

    Parameters
    ----------
    chemical1: Molable
        the first chemical structure, can be a SMILES string or an RDKit Mol object
        MUST be the same chemical as the one used to generate the assignment
    chemical2: Molable
        the second chemical structure, can be a SMILES string or an RDKit Mol object
        MUST be the same chemical as the one used to generate the assignment
    assignment: tuple[np.ndarray, np.ndarray]
        the assignment of two chemical structures, mapping nodes from one to the other
        returned by the approximate GED calculator if "return_assignment" is True
    save_path: Optional[PathLike], default=None
        path to save the plot to, if None, the plot will not be saved
    show: bool, default=True
        show the plot after generating it, if False, the plot will not be shown

    Returns
    -------
    Figure
        the matplotlib Figure object containing the plot of the assignment
    """
    g1 = mol_to_nx(to_mol(chemical1, fail_on_error=True))
    g2 = mol_to_nx(to_mol(chemical2, fail_on_error=True))

    g_comp = nx.disjoint_union_all([g1, g2])

    # Generate positions for each graph separately
    pos1 = nx.spring_layout(g1)
    pos2 = nx.spring_layout(g2)

    # Calculate centers and max distances for each graph
    center1 = np.mean(list(pos1.values()), axis=0)
    max_pos1 = np.max(np.abs(np.array(list(pos1.values())) - center1))

    center2 = np.mean(list(pos2.values()), axis=0)
    max_pos2 = np.max(np.abs(np.array(list(pos2.values())) - center2))

    # Combine positions, shifting the second graph to the right
    position = {}
    for node, pos in pos1.items():
        position[node] = pos

    shift = 2 * (np.max([max_pos1, max_pos2]) + 0.5)
    for node, pos in pos2.items():
        position[node + len(g1)] = pos.copy()
        position[node + len(g1)][0] += shift

    ag1, ag2 = assignment

    edgelist = []
    nodelist_ins = []
    nodelist_del = []
    for i1, i2 in zip(ag1, ag2):
        if i1 < len(g1):
            if i2 < len(g2):
                # Substitution
                g_comp.add_edge(i1, i2 + len(g1))
                edgelist += [(i1, i2 + len(g1))]
            else:
                # Deletion
                nodelist_del += [i1]
        else:
            # Insertion
            nodelist_ins += [i2 + len(g1)]

    center = np.mean(list(position.values()), axis=0)
    max_pos = np.max(np.abs(np.array(list(position.values())) - center), axis=0)

    fig = plt.figure()

    nx.draw_networkx_nodes(
        g_comp,
        position,
        nodelist=[item for item in g_comp.nodes if item not in nodelist_ins + nodelist_del],
        node_color="black",
        node_size=200,
    )

    nx.draw_networkx_edges(
        g_comp, position, edgelist=[item for item in g_comp.edges if item not in edgelist]
    )

    nx.draw_networkx_edges(
        g_comp, position, edgelist=edgelist, width=3, alpha=0.5, edge_color="b", style="dashed"
    )

    nx.draw_networkx_nodes(
        g_comp, position, nodelist=nodelist_ins, node_color="g", node_size=500, alpha=0.8
    )

    nx.draw_networkx_nodes(
        g_comp, position, nodelist=nodelist_del, node_color="r", node_size=500, alpha=0.8
    )

    plt.ylim([center[1] - max_pos[1] - 0.5, center[1] + max_pos[1] + 0.5])
    plt.xlim([center[0] - max_pos[0] - 0.5, center[0] + max_pos[0] + 0.5])
    plt.axis("off")

    if save_path is not None:
        plt.savefig(save_path)

    if show:
        plt.show()

    return fig
